"""
Returns the canonical retriangulation of a snappy manifold (for which
.canonize() has already been called) with shape field Q(sqrt(-3)). The
result is returned as a string.

It uses exact arithmetic to determine which faces of the proto-canonical
triangulations are opaque or transparent.

It also uses exact arithmetic and sage interval arithmetics to verify
hyperbolicity (exact arithmetics to verify that the edge equations are
fulfilled and cusps are complete, interval arithmetics to ceritify
positivity of the imaginary part of the shapes and the logarithmic lift
of the edge equations).

Relies on SnapPy and Sage. For testing, also on regina.

Use regina to generate to SnapPy Manifolds:

>>> from regina import NTriangulation
>>> from snappy import Manifold
>>> M=Manifold(NTriangulation.fromIsoSig('gLLPQccdfeefqjsqqjj').snapPea())
>>> N=Manifold(NTriangulation.fromIsoSig('fLLQcbdeedemnamjp').snapPea())

Take the proto-canonical triangulations

>>> M.canonize()
>>> N.canonize()

The first example has only tetrahedra in the canonical cell decomposition,
so all faces are opaue.

>>> CuspedManifold(M).find_opaque_faces()
[True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True, True]

We can now compute the canonical retriangulation (as string)
>>> s = canonical_retriangulation(M)

And check by taking the isomorphism signature
>>> NTriangulation.fromSnapPea(s).isoSig()
'KLLvPvMzzvLvLMPLwPQzQQQQQccehgkkmrypqvsEzFvDwxGzJFIHBIGHIJEEGJqfqfoaaaaaoqfoqfaoqooaoqaqafaqqaaofof'

The second example has non-tetrahedral cells in the canonical cell decompositon
so some faces are transparent.

>>> False in CuspedManifold(N).find_opaque_faces()
True

Check the result of the canonical retriangulation
>>> NTriangulation.fromSnapPea(canonical_retriangulation(N)).isoSig()
'mvvLALQQQhfghjjlilkjklaaaaaffffffff'

Also check for non-orientable manifold
>>> K = Manifold(NTriangulation.fromIsoSig('fLLQcaceeeddronwp').snapPea())
>>> K.canonize()
>>> NTriangulation.fromSnapPea(canonical_retriangulation(K)).isoSig()
'mLLzPLMMQcdeghigkjkllloaoaaaoaaooao'
"""

# This code is copyrighted by Nathan Dunfield, Neil Hoffman, and Joan Licata
# and released under the GNU GPL version 2 or (at your option) any later version.

# Rewritten to use exact arithmetic and emit the canonical retriangulation
# Matthias Goerner, 11/01/14

import snappy
import snappy.snap.t3mlite as t3m
from snappy.snap.t3mlite.simplex import *
import copy

from sage.rings.complex_interval_field import ComplexIntervalField
from sage.rings.real_mpfi import RealIntervalField

from fractions import Fraction
import sys
import math

RIF = RealIntervalField()
CIF = ComplexIntervalField()

class SquareRootCombination:
    """
    Represents a Q-linear combination of sqrt's of distinct positive
    square-free integers.
    Internally, c_1 * sqrt(r_1) + ... + c_n * sqrt(r_n) is represented by an
    array [(r_1, c_1), ..., (r_n, c_n)] such that r_i are ascending. The r_i
    are integers and the c_i integers or Fraction's.
    """

    @staticmethod
    def square_free(x):
        """
        Returns pair (t, y) such that x = t * y^2 and t is square-free.
        """

        y = 1
        i = 2
        # Not an efficient algorithm, but good enough for our purposes.
        while i * i <= x:
            
            if x % (i * i) == 0:
                x /= i * i
                y *= i
            else:
                i += 1
        
        return x, y

    def __init__(self, entries = []):
        """
        Constructs a Q-linear combination of square roots by
        normalizing the given entries [(root, coefficient), ...]
        """
        
        d = {}
        for root, coefficient in entries:
            square_free, extra_coefficient = (
                SquareRootCombination.square_free(root))
            d[square_free] = (
                d.get(square_free, 0) + coefficient * extra_coefficient)
        
        self._entries = sorted([(k,v) for (k,v) in d.items() if v])

    @staticmethod
    def Zero():
        return SquareRootCombination()

    @staticmethod
    def One():
        return SquareRootCombination([(1,1)])

    @staticmethod
    def Two():
        return SquareRootCombination([(1,2)])

    @staticmethod
    def Four():
        return SquareRootCombination([(1,4)])

    @staticmethod
    def SqrtThree():
        return SquareRootCombination([(3,1)])

    @staticmethod
    def guess_from_float(root, f):
        """
        Given an integer root and an object that can be converted to float f,
        tries to guess a representation of f as p/q * sqrt(root).
        """

        coeff = Fraction(float(f) / math.sqrt(root)).limit_denominator(10000)
        return SquareRootCombination([(root, coeff)])

    def __add__(self, other):
        return SquareRootCombination(self._entries + other._entries)

    def __neg__(self):
        return SquareRootCombination([(r, -c) for r, c in self._entries])

    def __sub__(self, other):
        return self + (-other)

    def __mul__(self, other):
        return SquareRootCombination(
            [(r1*r2, c1*c2) for r1, c1 in self._entries
                            for r2, c2 in other._entries])

    def __div__(self, other):

        assert len(other._entries) > 0, "Division by zero not allowed"

        if len(other._entries) == 1:
            root, coefficient = other._entries[0]
            inv = SquareRootCombination([ (root,
                                           Fraction(1) / coefficient / root) ])
            return self * inv

        p = 2
        while True:
            divs = [(r, 2 * c) for r, c in other._entries if r % p == 0]
            if divs:
                f = other - SquareRootCombination(divs)
                numerator = self * f
                denominator = other * f
                return numerator / denominator
            p += 1
    
    def __str__(self):
        if not self._entries:
            return '0'
        return '+'.join(['(%s * sqrt(%d))' % (c, r) for (r, c) in self._entries])

    def __repr__(self):
        return 'SquareRootCombination(%r)' % self._entries

    def sqrt(self):
        err_msg = "Only square roots of rational numbers are supported"

        if len(self._entries) == 0:
            return SquareRootCombination([])
        assert len(self._entries) == 1, err_msg
        root, coefficient = self._entries[0]
        assert root == 1, err_msg
        assert coefficient > 0, err_msg
        if isinstance(coefficient, int):
            return SquareRootCombination([(coefficient, 1)])
        return (
            SquareRootCombination([(coefficient.numerator, 1)]) /
            SquareRootCombination([(coefficient.denominator, 1)]))

    def __eq__(self, other):
        return self._entries == other._entries

    def evaluate(self):
        """
        Return an interval containing the true value.
        """
        return sum([ RIF(r).sqrt() * RIF(c.numerator) / RIF(c.denominator)
                     for r, c in self._entries])

class ComplexSquareRootCombination:
    """
    Represents a + b * i where a and b are Q-linear combinations of
    square roots of distinct square-free integers.

    This is implemented using SquareRootCombination objects stored
    under "real" and "imag".
    """


    def __init__(self, real, imag):
        """
        Constructs a + b * i given two SquareRootCombination objects.
        """

        self.real = real
        self.imag = imag

    @staticmethod
    def One():
        return ComplexSquareRootCombination(
            SquareRootCombination.One(), SquareRootCombination.Zero())

    def __repr__(self):
        return 'ComplexSquareRootCombination(%r, %r)' % (
            self.real, self.imag)

    def __abs__(self):
        return (self.real * self.real + self.imag * self.imag).sqrt()

    def __add__(self, other):
        return ComplexSquareRootCombination(
            self.real + other.real,
            self.imag + other.imag)

    def __neg__(self):
        return ComplexSquareRootCombination(
            -self.real, -self.imag)

    def __sub__(self, other):
        return self + (-other)

    def __mul__(self, other):
        return ComplexSquareRootCombination(
            self.real * other.real - self.imag * other.imag,
            self.real * other.imag + self.imag * other.real)

    def conjugate(self):
        return ComplexSquareRootCombination(
            self.real, -self.imag)

    def __div__(self, other):
        otherConj = other.conjugate()
        denom = (other * otherConj).real
        num = self * otherConj
        return ComplexSquareRootCombination(
            num.real / denom, num.imag /denom)

    def __eq__(self, other):
        return self.real == other.real and self.imag == other.imag

    def evaluate(self):
        """
        Returns a complex interval returning the true value.
        """

        return CIF(self.real.evaluate(), self.imag.evaluate())

#--- Some additional methods for t3m.Tetrahedron ----

def add_edge_parameters(self, shape):
    One = ComplexSquareRootCombination.One()

    z   = ComplexSquareRootCombination(
        SquareRootCombination.guess_from_float(1, shape.real()),
        SquareRootCombination.guess_from_float(3, shape.imag()))
    zp  = One / (One - z)
    zpp = (z - One) / z

    self.edge_params = {
        E01:z, E23:z, E02:zp, E13:zp, E03:zpp, E12:zpp}

def tilt(self, v):
    "The tilt of the face of the tetrahedron opposite the vertex v."
    ans = SquareRootCombination.Zero()
    for w in ZeroSubsimplices:
        if v == w:
            c_w = SquareRootCombination.One()
        else:
            z = self.edge_params[v | w]
            c_w = -z.real/abs(z)
        R_w = self.horotriangles[w].circumradius

        ans += c_w*R_w
    return ans

t3m.Tetrahedron.add_edge_parameters = add_edge_parameters
t3m.Tetrahedron.tilt = tilt

#-------- t3m helper code --------

FacesAnticlockwiseAroundVertices = {
    V0 : (F1, F2, F3),
    V1 : (F0, F3, F2), 
    V2 : (F0, F1, F3),
    V3 : (F0, F2, F1)}

def glued_to(tetrahedron, face):
    """
    Returns (other tet, other face).
    """
    return tetrahedron.Neighbor[face], tetrahedron.Gluing[face].image(face)

def tets_and_vertices_of_cusp(cusp):
    return [(corner.Tetrahedron, corner.Subsimplex) for corner in cusp.Corners]

#-------- Main classes --------

class HoroTriangle:
    """
    A horosphere cross section in the corner of an ideal tetrahedron.
    The sides of the triangle correspond to faces of the tetrahedron.
    """
    def __init__(self, tet, vertex, known_side, length_of_side):
        sides = FacesAnticlockwiseAroundVertices[vertex]
        left, center, right = HoroTriangle._make_middle(sides, known_side)
        z_l = tet.edge_params[left & center]
        z_r = tet.edge_params[center & right]
        L = length_of_side
        self.lengths = {center:L, left:abs(z_l)*L, right:L/abs(z_r)}
        a, b, c = self.lengths.values()

        self.area = L * L * z_l.imag / SquareRootCombination.Two()
        # Below is the usual formula for circumradius combined with 
        # Heron's formula.
        self.circumradius = (a * b * c /
                             (SquareRootCombination.Four() * self.area))

    def rescale(self, t):
        "Rescales the triangle by a Euclidean dilation"
        for face, length in self.lengths.items():
            self.lengths[face] = t*length
        self.circumradius = t*self.circumradius
        self.area = t*t*self.area

    @staticmethod
    def _make_middle((a,b,c), x):
        "Cyclically rotate (a,b,c) so that x is the middle entry"
        if x == a:
            return (c,a,b)
        elif x == b:
            return (a,b,c)
        elif x == c:
            return (b,c,a)


class CuspedManifold(t3m.Mcomplex):
    """
    A t3m triangulation built from a SnapPy manifold with
    a hyperbolic structure specified by HIKMOT's guaranteed
    shape intervals.
    """
    def __init__(self, manifold):
        t3m.Mcomplex.__init__(self, manifold)
        self.snappy_manifold = manifold
        self.add_shapes()
        for T in self.Tetrahedra:
            T.horotriangles = {V0:None, V1:None, V2:None, V3:None}
        self.add_cusp_cross_sections()

        # Some sanity checks.  
        self.check_edge_equations()
        self.check_cusp_cross_sections()

    def add_shapes(self):
        
        shapes = [ shape_dict['rect']
                   for shape_dict
                   in self.snappy_manifold.tetrahedra_shapes() ]

        for T, z in zip(self.Tetrahedra, shapes):
            add_edge_parameters(T, z)

    def add_cusp_cross_sections(self):
        for cusp in self.Vertices:
            self.add_one_cusp_cross_section(cusp)

    def add_one_cusp_cross_section(self, cusp):
        "Build a cusp cross section as described in Section 3.6 of the paper"
        tet0, vert0 = tets_and_vertices_of_cusp(cusp)[0]
        face0 = FacesAnticlockwiseAroundVertices[vert0][0]
        tet0.horotriangles[vert0] = HoroTriangle(
            tet0, vert0, face0,
            SquareRootCombination.One())
        active = [(tet0, vert0)]
        while active:
            tet0, vert0 = active.pop()
            for face0 in FacesAnticlockwiseAroundVertices[vert0]:
                tet1, face1 = glued_to(tet0, face0)
                vert1 = tet0.Gluing[face0].image(vert0)
                if tet1.horotriangles[vert1] is None:
                    tet1.horotriangles[vert1] = HoroTriangle(tet1, vert1, face1,
                                tet0.horotriangles[vert0].lengths[face0])
                    active.append( (tet1, vert1) )

    def _get_cusp(self, cusp):
        """
        Helper method so the user can specify a cusp by its index as well
        the actual t3m.Vertex.
        """
        if not isinstance(cusp, t3m.Vertex):
            cusp = self.Vertices[cusp]
        return cusp
            
    def cusp_area(self, cusp):
        cusp = self._get_cusp(cusp)
        area = SquareRootCombination.Zero()
        for T, V in tets_and_vertices_of_cusp(cusp):
            area += T.horotriangles[V].area
        return area

    def rescale_cusp(self, cusp, scale):
        cusp = self._get_cusp(cusp)
        for T, V in tets_and_vertices_of_cusp(cusp):
            T.horotriangles[V].rescale(scale)
            
    def normalize_cusp(self, cusp):
        """
        Rescale cusp to have area sqrt(3). This choice ensures that
        all tilts are again Q-linear combinations of square roots
        of integers.
        """
        cusp = self._get_cusp(cusp)

        target_area = SquareRootCombination.SqrtThree()

        area = self.cusp_area(cusp)
        ratio = (target_area/area).sqrt()
        self.rescale_cusp(cusp, ratio)

    def LHS_of_convexity_equations(self):
        """
        For each face in the triangulation, return a quantity which is < 0
        if and only if the corresponding pair of tetrahedra are strictly
        convex.
        """
        ans = []
        for tet0 in self.Tetrahedra:
            for vert0 in ZeroSubsimplices:
                tet1, face1 = glued_to(tet0, comp(vert0))
                ans.append(tet0.tilt(vert0) + tet1.tilt(comp(face1)))
        return ans

    def find_opaque_faces(self):
        """
        Returns a list of bools indicating whether a face of a tetrahedron
        of the given proto-canonical triangulation is opaqure.
        The list is of the form
        [ face0_tet0, face1_tet0, face2_tet0, face3_tet0, face0_tet1, ...]
        """
        num_cusps = len(self.Vertices)
        for i in range(num_cusps):
            self.normalize_cusp(i)
        
        tilts = self.LHS_of_convexity_equations()
        result = []
        for tilt in tilts:
            # Face is transparent when tilt is exactly 0
            if tilt == SquareRootCombination.Zero():
                result.append(False)
            # Use interval aritmetic to certify tilt is negative
            elif tilt.evaluate() < 0:
                result.append(True)
            else:
                # We failed
                raise Exception(
                    "Could not certify proto-canonical triangulation")
                
        return result

    #------- All remaining methods are just sanity and regression checks; safety first! -----------
    
    def check_cusp_cross_sections(self):
        """
        Sanity check: do all pairs of adjacent triangles have the same
        edge lengths?
        """
        errors = []
        for tet0 in self.Tetrahedra:
            for vert0 in ZeroSubsimplices:
                for face0 in FacesAnticlockwiseAroundVertices[vert0]:
                    tet1, face1 = glued_to(tet0, face0)
                    vert1 = tet0.Gluing[face0].image(vert0)
                    side0 = tet0.horotriangles[vert0].lengths[face0]
                    side1 = tet1.horotriangles[vert1].lengths[face1]

                    assert side0 == side1

        return errors

    @staticmethod
    def first_edge_embedding(edge):
        """
        For a given t3m.Edge edge, return an edge embedding similar
        to regina, that is a pair (tetrahedron, permutation)
        such that vertex 0 and 1 of the tetrahedron span the edge.
        """

        corner = edge.Corners[0]
        
        for p in [(0,1,2,3),(0,2,1,3),(0,3,1,2),
                  (1,2,0,3),(1,3,0,2),(2,3,0,1)]:
            perm = t3m.Perm4(p)
            if corner.Subsimplex == perm.image(3):
                return (corner.Tetrahedron, perm)

    @staticmethod
    def next_edge_embedding(tet, perm):
        """
        Given a pair (tet, perm) of the form above, give the next
        edge embedding.
        """
        face = 15 - (1 << perm[2])
        return tet.Neighbor[face], tet.Gluing[face] * perm * t3m.Perm4((0,1,3,2))

    def check_edge_equations(self):
        """
        Check edge equation using exact arithmetic. Check logarithmic lifts
        using interval arithmetics.
        """

        # For each edge
        for edge in self.Edges:

            # The exact value when evaluating the edge equation
            exact       = ComplexSquareRootCombination.One()
            # The complex interval arithmetic value of the logarithmic
            # version of the edge equation.
            numeric_log = CIF(0)
            
            # Iterate through edge embeddings
            order = len(edge.Corners)
            tet, perm = CuspedManifold.first_edge_embedding(edge)
            for i in range(order):
                
                # Get the shape for this edge embedding
                subsimplex = perm.image(3)
                this_exact = tet.edge_params[subsimplex] 

                # Figure out the orientation of this tetrahedron
                # with respect to the edge, apply conjugate inverse
                # if differ
                if perm.sign():
                    this_exact = (
                        ComplexSquareRootCombination.One() /
                        this_exact.conjugate())
                    
                # Accumulate shapes of the edge exactly
                exact = exact * this_exact

                # Convert to numerical value
                this_numeric = this_exact.evaluate()
                # Certify positive imaginary part of shape
                assert this_numeric.imag > 0, (
                    "Could not certify postive imaginary part of shape")
                # Take logarithm and accumulate
                numeric_log += this_numeric.log()

                # Find next edge embedding
                tet, perm = CuspedManifold.next_edge_embedding(tet, perm)

            # Check that edge equations is exactly one
            assert exact == ComplexSquareRootCombination.One(), exact
            # And logarithmic lift is close to 2 pi i, epsilon
            # is small enough to ensure the correct branch.
            assert abs(numeric_log - CIF.pi() * CIF(0,2)) < 1e-7

#--- convenience functions ------

def canonical_retriangulation(manifold):
    """
    Takes as input a snappy manifold with shape field Q(sqrt(-3)) for which
    .canonize() has been applied.
    Returns the canonical retriangulation as string if cells are
    non-tetrahedral, otherwise, just return the input manifold as string.
    """
    
    if not hasattr(manifold, "_canonical_retriangulation_to_string"):
        raise Exception("This requires a version of SnapPy with the patch "
                        "SnapPyExposeCanonicalRetriangulation applied, "
                        "see README.txt for instructions.")

    opaque_faces = CuspedManifold(manifold).find_opaque_faces()

    # If all cells are tetrahedral, don't perform canonical retriangulation
    if not False in opaque_faces:
        return manifold._to_string()
    
    return manifold._canonical_retriangulation_to_string(opaque_faces)

if __name__ == '__main__':
   import doctest
   doctest.testmod()
