import sys
from regina import NTriangulation

from readDataFiles import parse_text
from morphismMethods import find_morphisms

def filter_morphisms_that_factor(list_of_manifolds, is_covering):
    """
    Given a list of manifold names and the is covering relationship
    A -> B encoded as described below for is_covering,
    return the result when removing all B's such that A is in the list.
    """

    covered_by_list = set()
    for manifold in list_of_manifolds:
        covered_by_list |= set(is_covering[manifold])

    return [ manifold
             for manifold in list_of_manifolds
             if not manifold in covered_by_list ]
    

def read_input_files(filenames):
    """
    Reads the given files and concatenates the result.
    """
    return sum(
        [ parse_text(open(filename).read())
          for filename in filenames ], [])

if __name__ == '__main__':

    # Read input files and concatenate
    named_list_of_list_of_iso_sigs = read_input_files(sys.argv[1:])
    # Sort by number of tetrahedra
    named_list_of_list_of_iso_sigs.sort(
        key = lambda t: NTriangulation.isoSigComponentSize(t[1][0]))

    # The is_covering relationship.
    # It is a dictionary where key is the name of a triangulation
    # and value is a list of names.
    # This encodes that key is a covering of the manifolds in value.
    # Covering maps of degree 1 (e.g., automorphisms) are skipped.
    is_covering = {}

    # named_list_of_list_of_iso_sigs is a list of triples
    # (name, listOfIsomoSig, isometrySig)
    # name is something like otet05_0003
    # The full name for the i-th triangulation in listOfIsomoSig
    # is otet05_0003#i
    # We don't care about isometrySig

    # Iterate through all triangulations
    # First loop iterates through isometry types
    for src_name, src_iso_sigs, d in named_list_of_list_of_iso_sigs:
        for src_index, src_iso_sig in enumerate(src_iso_sigs):

            # src_name is like "otet05_0003"
            # src_full_name is like "otet05_0003#1"
            src_full_name = src_name + '#%d' % src_index
            
            # Number of tetrahedra
            src_iso_sig_size = NTriangulation.isoSigComponentSize(src_iso_sig)

            # Initialize the is_covering relationship
            is_covering[src_full_name] = []

            # Inner loop through all triangulations
            for dst_name, dst_iso_sigs, d2 in named_list_of_list_of_iso_sigs:
                for dst_index, dst_iso_sig in enumerate(dst_iso_sigs):
                    
                    # Number of tetrahedra
                    dst_iso_sig_size = (
                        NTriangulation.isoSigComponentSize(dst_iso_sig))

                    # Get full name
                    dst_full_name = dst_name + '#%d' % dst_index

                    # Skip covering maps of degree 1
                    # Skip maps where number of tet's don't divide
                    if (src_iso_sig_size > dst_iso_sig_size and 
                        src_iso_sig_size % dst_iso_sig_size == 0):
                        
                        # Create the regina triangulations
                        src = NTriangulation.fromIsoSig(src_iso_sig)
                        dst = NTriangulation.fromIsoSig(dst_iso_sig)

                        print >>sys.stderr, (
                            "Checking: %s -> %s" % (src_full_name,
                                                    dst_full_name))

                        # Check if there are morphisms
                        if find_morphisms(src, dst):
                            # Add to is_covering relationship
                            is_covering[src_full_name].append(dst_full_name)

            # If there are morphisms A->B->C, then
            # is_covering[A] contains both B and C, filter out C
            filtered = filter_morphisms_that_factor(
                is_covering[src_full_name], is_covering)

            # Output the morphisms
            for dst_full_name in filtered:
                print src_full_name, "->", dst_full_name#, ";"
            sys.stdout.flush()
    
